/**
 * netzond is a real-time charting/graphing applet
 * Copyright (C) 2007 johnrembo@zebra.lt
 * 
 * The source-code of netzond applet comes under GPL2 License  
 * 
 *@author       johnrembo@zebra.lt
 *@version      1.04
 *@since        1.0
 */

package netzond;

import java.applet.*;
import java.awt.*;
import java.net.*;
import java.io.*;
import java.util.*;
import java.lang.Math;
import java.lang.String;
import java.awt.AlphaComposite.*;
import java.awt.GradientPaint;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GradientPaint;
import java.awt.image.BufferedImage;
import java.awt.geom.*;
import java.math.*;
import java.text.*;

public class netzond extends Applet implements Runnable {
    //2D BUFFER
    BufferedImage buffer;
    Graphics2D bufimg;

    //DATA FETCH
    private double[] currentData,  maxCollVal,  minCollVal;
    private double[][] dataCollection;
    private int buff,  dataAmount,  collectionIndex;
    //OUTER REGION
    private int border,  lineSpacing;
    //AXIES
    private double southWestHorzMax,  southWestHorzMin,  southWestHorzDelta,  northEastHorzMax,  northEastHorzMin,  northEastHorzDelta,  globalMaxSW,  globalMaxNE;
    private boolean horzScaleNE,  horzScaleSW;
    //FONTS
    //  original font sizes
    private int fs1Unadopted,  fs2Unadopted,  fs3Unadopted;
    //  DPI adopted font sizes
    private int fontSize1,  fontSize2,  fontSize3;
    //  Font size scatter
    private int fs1Bias,  fs2Bias,  fs3Bias;
    private int fontPixelSize1,  fontPixelSize2,  fontPixelSize3;
    //TITLES
    private int graphTitleHeight,  graphTitleTopHeight,  graphTitleBotHeight,  graphTitleTopWidth,  graphTitleBotWidth,  axisTitleHeight,  axisTitleTopHeight,  axisTitleBotHeight,  axisTitleLeftHeight,  axisTitleRightHeight,  axisTitleTopWidth,  axisTitleBotWidth,  axisTitleLeftWidth,  axisTitleRightWidth;
    private String graphTitleTop,  graphTitleBot,  axisTitleTop,  axisTitleBot,  axisTitleLeft,  axisTitleRight;

    //legends and values
    private int legendAreaHeight,  legendCount,  legendSquareSize,  valuesHeight;
    private String[] legendValue;
    private int[] legendIndexes;

    //SCALES
    private String scaleTop,  scaleBot,  DATE_FORMAT, trlSymbNE,  tlrSymbSW, ldSymbNE, ldSymbSW;
    //k-kilo, M-mega, G-giga, T-tera, P-peta, E-egza, Z-zeta, Y-yota
    private String[] shrtctRight = {"k", "M", "G", "T", "P", "E", "Z", "Y"};
    //m-mili, µ-mikro, n-nano, p-piko, f-femto, a-ato, z-zepto, y-yokto
    private String[] shrtctLeft = {"m", "µ", "n", "p", "f", "a", "z", "y"};
    private BigDecimal[][] baseR,  baseL;
    private int[] valuesLeft,  valuesBot;
    private int X,  Y,  prevX,  i,  j;

    //LEGEND
    private String[] legendStatisticsMin,  legendStatisticsAvg,  legendStatisticsMax;
    NumberFormat formatterE = new DecimalFormat();
    String F="";
    String Fexp = "###E0";
                
    //UNITS
    private String unitsSW,  unitsNE,  dimensionSW,  dimensionNE,  dimensionOpen,  dimensionClose,  direction,  tempString;
    private int dimensionFactor,  dimensionFactorSW,  dimensionFactorNE;
    private BigDecimal factorSW = BigDecimal.ONE;
    private BigDecimal factorNE = BigDecimal.ONE;

    //COLORS
    private Color[] Palette;
    private int[] transIndex,  colorIndex;
    private boolean transparent = true;
    private boolean coloredLegendText,  retainAxisValuesColors;
    private Color currCol;

    //CHART
    private int marginTop,  marginBot,  marginLeft,  marginRight,  graphWidth,  graphHeight;
    private int step,  chartInnerMarginHorz,  chartInnerMarginVert,  majVertGridDensity,  minVertGridDensity,  majHorzGridDensity,  minHorzGridDensity,  majVertGridLines,  majHorzGridLines,  minHorzGridLines,  valuesAreaDataWidthSW,  valuesAreaDataWidthSwPix,valuesAreaDataWidthNE,  valuesAreaDataWidthNePix,  valuesAreaTimeWidth,  valuesAreaTopHeight,  valuesAreaBotHeight,  roundScale,  roundScaleSW,  roundScaleNE;
    private BigDecimal delta,  count,  min,  max;
    private boolean shifted,  indicator;
    private boolean[] showValuesOnChart;
    int[][] xArr, yArr;

    //OTHER
    int screenRes, leftPart, refreshInterval;
    private String[] Way;
    private String tempValue,  axisValue,  fontName,  path;
    private Font font1,  font2,  font3;
    private FontMetrics fontMetrics;
    private Dimension d;
    public Thread graphThread = null;
    boolean antiAlias;
    BigDecimal zero = BigDecimal.ZERO;

    //LAYOUT
    private int inputDataCount,  displayBot,  displayTop,  legendSeparator,  legendsPerLine;
    private int[] southWest,  northEast,  prevYSW,  prevYNE;
    private int[][] lineColor;
    private String legendType,  indicatorType;
    private String[] graphType;
    private boolean rotatedTop,  rotatedBot,  fixedScaleMinNE,  fixedScaleMaxNE,  fixedScaleMinSW,  fixedScaleMaxSW,  trafficScaleSW,  trafficScaleNE,  trafficScale,  SWS,  NES;
    private boolean displayLeft = true;
    private boolean displayRight = true;
    private boolean[] lineSmothness,  legendDetails,  legendUnits,  currValue,  lineContour,  hideLine, traffScale;
    private double scaleMinNE,  scaleMaxNE,  scaleMinSW,  scaleMaxSW;
    private float indicatorWidth;

    //INPUT
    String[] swLineArray, neLineArray, SWParamsArray, NEParamsArray, dataParamsArray, outerParamsArray, innerParamsArray, colorsArray, objectColorsArray, lineTypes;
    private float[] lineWidths;
    private int inc = 0;
    
    //RESTRICTIONS
    private int lineParametersCount = 15;
    private int dataParamsCount = 5;
    private int SWParamsCount = 9;
    private int NEParamsCount = 9;
    private int outerParamsCount = 15;
    private int innerParamsCount = 19;
    //USER INPUT
    String dataParams, SWParams, NEParams, outerParams, innerParams, colorParams, objcolsParams, swString, neString;

    public void init() {
        String param = getParameter("dataParams") ; if (param != null) dataParams = param ; else throw new Error("dataParams must not be empty!");
        param = getParameter("SWParams") ; SWParams = param ;
        param = getParameter("NEParams") ; NEParams = param ;
        param = getParameter("outerParams") ; if (param != null) outerParams = param ; else throw new Error("outerParams must not be empty!");
        param = getParameter("innerParams") ; if (param != null) innerParams = param ; else throw new Error("innerParams must not be empty!");
        param = getParameter("colorParams") ; if (param != null) colorParams = param ; else throw new Error("colorParams must not be empty!");
        param = getParameter("objcolsParams") ; if (param != null) objcolsParams = param ; else throw new Error("objcolsParams must not be empty!");
        param = getParameter("swString") ; swString = param ;
        param = getParameter("neString") ; neString = param ;

        if (dataParams != null) {
            dataParamsArray = dataParams.split(",", -1);
            if (dataParamsArray.length != dataParamsCount) {
                System.out.println("incorrect dataParams count. Should be " + dataParamsCount + " instead of " + dataParamsArray.length);
            }
            inc = 0;
            /*1*/            scaleTop = dataParamsArray[inc];
            inc++;
            /*2*/            scaleBot = dataParamsArray[inc];
            inc++;
            /*3*/            DATE_FORMAT = dataParamsArray[inc];
            inc++;
            /*4*/            dimensionOpen = dataParamsArray[inc];
            inc++;
            /*5*/            dimensionClose = dataParamsArray[inc];
        } else {
            throw new Error("Empty dataParams string");
        }

        if (SWParams != null) {
            SWParamsArray = SWParams.split(",", -1);
            if (SWParamsArray.length != SWParamsCount) {
                System.out.println("incorrect SWParamsArray count. Should be " + SWParamsCount + " instead of " + SWParamsArray.length);
            }
            inc = 0;
            /*1*/            dimensionSW = SWParamsArray[inc];
            inc++;
            /*2*/            dimensionFactorSW = Integer.parseInt(SWParamsArray[inc]);
            inc++;
            /*3*/            roundScaleSW = Integer.parseInt(SWParamsArray[inc]);
            inc++;
            /*4*/            trafficScaleSW = Boolean.parseBoolean(SWParamsArray[inc]);
            inc++;
            /*5*/            fixedScaleMaxSW = Boolean.parseBoolean(SWParamsArray[inc]);
            inc++;
            /*6*/            scaleMaxSW = Double.parseDouble(SWParamsArray[inc]);
            inc++;
            /*7*/            fixedScaleMinSW = Boolean.parseBoolean(SWParamsArray[inc]);
            inc++;
            /*8*/            scaleMinSW = Double.parseDouble(SWParamsArray[inc]);
            inc++;
            /*9*/            valuesAreaDataWidthSW = Integer.parseInt(SWParamsArray[inc]);
            
            ldSymbSW = fill("","#",(valuesAreaDataWidthSW-roundScaleSW))+"."+fill("","#",roundScaleSW);
        }

        if (NEParams != null) {
            NEParamsArray = NEParams.split(",", -1);
            if (NEParamsArray.length != NEParamsCount) {
                System.out.println("incorrect NEParamsArray count. Should be " + NEParamsCount + " instead of " + NEParamsArray.length);
            }
            inc = 0;
            /*1*/            dimensionNE = NEParamsArray[inc];
            inc++;
            /*2*/            dimensionFactorNE = Integer.parseInt(NEParamsArray[inc]);
            inc++;
            /*3*/            roundScaleNE = Integer.parseInt(NEParamsArray[inc]);
            inc++;
            /*4*/            trafficScaleNE = Boolean.parseBoolean(NEParamsArray[inc]);
            inc++;
            /*5*/            fixedScaleMaxNE = Boolean.parseBoolean(NEParamsArray[inc]);
            inc++;
            /*6*/            scaleMaxNE = Double.parseDouble(NEParamsArray[inc]);
            inc++;
            /*7*/            fixedScaleMinNE = Boolean.parseBoolean(NEParamsArray[inc]);
            inc++;
            /*8*/            scaleMinNE = Double.parseDouble(NEParamsArray[inc]);
            inc++;
            /*9*/            valuesAreaDataWidthNE = Integer.parseInt(NEParamsArray[inc]);
            
            ldSymbNE = fill("","#",(valuesAreaDataWidthNE-roundScaleNE))+"."+fill("","#",roundScaleNE);
        }

        if (SWParams == null && NEParams == null) {
            throw new Error("SWParams or NEParams must not be empty");
        }

        if (outerParams != null) {
            outerParamsArray = outerParams.split(",", -1);
            if (outerParamsArray.length != outerParamsCount) {
                System.out.println("incorrect outerParams count. Should be " + outerParamsCount + " instead of " + outerParamsArray.length);
            }
            inc = 0;
            /*1*/            border = Integer.parseInt(outerParamsArray[inc]);
            inc++;
            /*2*/            lineSpacing = Integer.parseInt(outerParamsArray[inc]);
            inc++;
            /*3*/            fontName = outerParamsArray[inc];
            inc++;
            /*4*/            fs1Unadopted = Integer.parseInt(outerParamsArray[inc]);
            inc++;
            /*5*/            fs2Unadopted = Integer.parseInt(outerParamsArray[inc]);
            inc++;
            /*6*/            fs3Unadopted = Integer.parseInt(outerParamsArray[inc]);
            inc++;
            /*7*/            graphTitleTop = outerParamsArray[inc];
            inc++;
            /*8*/            graphTitleBot = outerParamsArray[inc];
            inc++;
            /*9*/            axisTitleTop = outerParamsArray[inc];
            inc++;
            /*10*/            axisTitleBot = outerParamsArray[inc];
            inc++;
            /*11*/            axisTitleLeft = outerParamsArray[inc];
            inc++;
            /*12*/            axisTitleRight = outerParamsArray[inc];
            inc++;
            /*13*/            indicator = Boolean.parseBoolean(outerParamsArray[inc]);
            inc++;
            /*14*/            refreshInterval = Integer.parseInt(outerParamsArray[inc]);
            inc++;
            /*15*/            path = outerParamsArray[inc];
        } else {
            throw new Error("Empty outerParams string");
        }

        if (innerParams != null) {
            innerParamsArray = innerParams.split(",", -1);
            if (innerParamsArray.length != innerParamsCount) {
                System.out.println("incorrect innerParams count. Should be " + innerParamsCount + " instead of " + innerParamsArray.length);
            }
            inc = 0;
            /*1*/            step = Integer.parseInt(innerParamsArray[inc]);
            inc++;
            /*2*/            chartInnerMarginHorz = Integer.parseInt(innerParamsArray[inc]);
            inc++;
            /*3*/            chartInnerMarginVert = Integer.parseInt(innerParamsArray[inc]);
            inc++;
            /*4*/            majVertGridDensity = Integer.parseInt(innerParamsArray[inc]);
            inc++;
            /*5*/            minVertGridDensity = Integer.parseInt(innerParamsArray[inc]);
            inc++;
            /*6*/            majHorzGridDensity = Integer.parseInt(innerParamsArray[inc]);
            inc++;
            /*7*/            minHorzGridDensity = Integer.parseInt(innerParamsArray[inc]);
            inc++;
            /*8*/            displayBot = Integer.parseInt(innerParamsArray[inc]);
            inc++;
            /*9*/            displayTop = Integer.parseInt(innerParamsArray[inc]);
            inc++;
            /*10*/            displayLeft = Boolean.parseBoolean(innerParamsArray[inc]);
            inc++;
            /*11*/            displayRight = Boolean.parseBoolean(innerParamsArray[inc]);
            inc++;
            /*12*/            rotatedBot = Boolean.parseBoolean(innerParamsArray[inc]);
            inc++;
            /*13*/            rotatedTop = Boolean.parseBoolean(innerParamsArray[inc]);
            inc++;
            /*14*/            indicatorType = innerParamsArray[inc];
            inc++;
            /*15*/            indicatorWidth = Float.parseFloat(innerParamsArray[inc]);
            inc++;
            /*16*/            legendSeparator = Integer.parseInt(innerParamsArray[inc]);
            inc++;
            /*17*/            legendsPerLine = Integer.parseInt(innerParamsArray[inc]);
            inc++;
            /*18*/            legendType = innerParamsArray[inc];
            inc++;
            /*19*/            retainAxisValuesColors = Boolean.parseBoolean(innerParamsArray[inc]);
        } else {
            throw new Error("Empty innerParams string");
        }

        //extract separate lines with their parameters to array
        String[] swPlotArray = new String[0];
        String[] nePlotArray = new String[0];
        if (swString != null) {
            swPlotArray = swString.split(";");
            southWest = new int[swPlotArray.length];
            inputDataCount += southWest.length;
            if (neString == null) {
                displayLeft = false;
            }
            SWS = true;
        }

        if (neString != null) {
            nePlotArray = neString.split(";");
            northEast = new int[nePlotArray.length];
            inputDataCount += northEast.length;
            if (swString == null) {
                displayRight = false;
            }
            NES = true;
        }
        if (!SWS && !NES) {
            throw new Error("swString [" + swString + "] or neString [" + neString + "] must not be empty");
        }

        swLineArray = new String[lineParametersCount];
        neLineArray = new String[lineParametersCount];

        if (southWest != null) {
            prevYSW = new int[southWest.length];
        }
        if (northEast != null) {
            prevYNE = new int[northEast.length];
        }

        //arrays containing per-line values
        //inputDataCount = southWest.length + northEast.length;
        /*1*/        legendValue = new String[inputDataCount];
        /*2*/        graphType = new String[inputDataCount];
        /*3*/        lineColor = new int[inputDataCount][2];
        /*4*/        transIndex = new int[inputDataCount];
        /*5*/        lineSmothness = new boolean[inputDataCount];
        /*6*/        legendDetails = new boolean[inputDataCount];
        /*7*/        legendUnits = new boolean[inputDataCount];
        /*8*/        currValue = new boolean[inputDataCount];
        /*9*/        lineTypes = new String[inputDataCount];
        /*10*/        lineWidths = new float[inputDataCount];
        /*11*/        showValuesOnChart = new boolean[inputDataCount];
        /*12*/        lineContour = new boolean[inputDataCount];
        /*13*/        hideLine = new boolean[inputDataCount];
        /*14*/        boolean createStatisticsArray = false;
        traffScale = new boolean[inputDataCount];
        
        //Color values array
        String[] colorsArray = colorParams.split(";");

        //Object colors array
        String[] tempArray = objcolsParams.split(",");
        colorIndex = new int[tempArray.length];
        for (i = 0; i < colorIndex.length; i++) {
            colorIndex[i] = Integer.parseInt(tempArray[i]);
        }

        if (SWS) {
            for (i = 0; i < swPlotArray.length; i++) {
                inc = 0;
                //extract certain line parrameters to array        
                swLineArray = swPlotArray[i].split(",", -1);
                if (swLineArray.length != lineParametersCount) {
                    System.out.println("SW: not enough parameters for line " + i);
                    break;
                }
                southWest[i] = Integer.parseInt(swLineArray[inc]);
                inc++; //southWest[1]=-1;
                if (southWest[i] < 0) {
                    break;
                }
                /*1*/                graphType[southWest[i]] = swLineArray[inc];
                inc++;
                /*2*/                lineColor[southWest[i]][0] = Integer.parseInt(swLineArray[inc]);
                inc++;
                /*3*/                lineColor[southWest[i]][1] = Integer.parseInt(swLineArray[inc]);
                inc++;
                /*4*/                transIndex[southWest[i]] = Integer.parseInt(swLineArray[inc]);
                inc++;
                /*5*/                lineSmothness[southWest[i]] = Boolean.parseBoolean(swLineArray[inc]);
                inc++;
                /*6*/                legendDetails[southWest[i]] = Boolean.parseBoolean(swLineArray[inc]);
                inc++;
                if (legendDetails[southWest[i]]) {
                    createStatisticsArray = true;
                }
                /*7*/                legendValue[southWest[i]] = swLineArray[inc];
                inc++;
                /*8*/                legendUnits[southWest[i]] = Boolean.parseBoolean(swLineArray[inc]);
                inc++;
                /*9*/                currValue[southWest[i]] = Boolean.parseBoolean(swLineArray[inc]);
                inc++;
                /*10*/                lineTypes[southWest[i]] = swLineArray[inc];
                inc++;
                /*11*/                lineWidths[southWest[i]] = Float.parseFloat(swLineArray[inc]);
                inc++;
                /*12*/                showValuesOnChart[southWest[i]] = Boolean.parseBoolean(swLineArray[inc]);
                inc++;
                /*13*/                lineContour[southWest[i]] = Boolean.parseBoolean(swLineArray[inc]);
                inc++;
                /*14*/                hideLine[southWest[i]] = Boolean.parseBoolean(swLineArray[inc]);
            }
        }

        if (NES) {
            for (i = 0; i < nePlotArray.length; i++) {
                inc = 0;
                neLineArray = nePlotArray[i].split(",", -1);
                if (neLineArray.length != lineParametersCount) {
                    System.out.println("NE: not enough parameters for line " + i);
                    break;
                }
                northEast[i] = Integer.parseInt(neLineArray[inc]);
                inc++; //northEast[1]=-1;
                if (northEast[i] < 0) {
                    break;
                }
                /*1*/                graphType[northEast[i]] = neLineArray[inc];
                inc++;
                /*2*/                lineColor[northEast[i]][0] = Integer.parseInt(neLineArray[inc]);
                inc++;
                /*3*/                lineColor[northEast[i]][1] = Integer.parseInt(neLineArray[inc]);
                inc++;
                /*4*/                transIndex[northEast[i]] = Integer.parseInt(neLineArray[inc]);
                inc++;
                /*5*/                lineSmothness[northEast[i]] = Boolean.parseBoolean(neLineArray[inc]);
                inc++;
                /*6*/                legendDetails[northEast[i]] = Boolean.parseBoolean(neLineArray[inc]);
                inc++;
                if (legendDetails[northEast[i]]) {
                    createStatisticsArray = true;
                }
                /*7*/                legendValue[northEast[i]] = neLineArray[inc];
                inc++;
                /*8*/                legendUnits[northEast[i]] = Boolean.parseBoolean(neLineArray[inc]);
                inc++;
                /*9*/                currValue[northEast[i]] = Boolean.parseBoolean(neLineArray[inc]);
                inc++;
                /*10*/                lineTypes[northEast[i]] = neLineArray[inc];
                inc++;
                /*11*/                lineWidths[northEast[i]] = Float.parseFloat(neLineArray[inc]);
                inc++;
                /*12*/                showValuesOnChart[northEast[i]] = Boolean.parseBoolean(neLineArray[inc]);
                inc++;
                /*13*/                lineContour[northEast[i]] = Boolean.parseBoolean(neLineArray[inc]);
                inc++;
                /*14*/                hideLine[northEast[i]] = Boolean.parseBoolean(neLineArray[inc]);
            }
        }

        Palette = new Color[colorsArray.length];
        String[] temPal;
        for (i = 0; i < Palette.length; i++) {
            temPal = colorsArray[i].split(",");
            if (temPal.length != 3) {
                System.out.println("Incorrect " + i + "-th RGB color definition. Must contain 3 values (from 0 to 255 each) separated by commas.");
            }
            Palette[i] = new Color(Integer.parseInt(temPal[0]), Integer.parseInt(temPal[1]), Integer.parseInt(temPal[2]));
        }

        screenRes = Toolkit.getDefaultToolkit().getScreenResolution();

        if (createStatisticsArray) {
            legendStatisticsMin = new String[inputDataCount];
            legendStatisticsAvg = new String[inputDataCount];
            legendStatisticsMax = new String[inputDataCount];
        }

        fontSize3 = (int) Math.round(fs3Unadopted * screenRes / 72.0);
        fontSize2 = (int) Math.round(fs2Unadopted * screenRes / 72.0);
        fontSize1 = (int) Math.round(fs1Unadopted * screenRes / 72.0);

        fontPixelSize1 = fs1Unadopted * screenRes / 72;
        fontPixelSize2 = fs2Unadopted * screenRes / 72;
        fontPixelSize3 = fs3Unadopted * screenRes / 72;

        fs1Bias = fontSize1 - fs1Unadopted;
        fs2Bias = fontSize2 - fs2Unadopted;
        fs3Bias = fontSize3 - fs3Unadopted;

        font3 = new Font(fontName, Font.PLAIN, fontSize3);
        font2 = new Font(fontName, Font.PLAIN, fontSize2);
        font1 = new Font(fontName, Font.PLAIN, fontSize1);

        fontMetrics = getFontMetrics(font3);
        graphTitleHeight = fontPixelSize3 + lineSpacing;
        if (!graphTitleTop.equals("")) {
            graphTitleTopHeight = graphTitleHeight;
            graphTitleTopWidth = fontMetrics.stringWidth(graphTitleTop);
        }
        if (!graphTitleBot.equals("")) {
            graphTitleBotHeight = graphTitleHeight;
            graphTitleBotWidth = fontMetrics.stringWidth(graphTitleBot);
        }

        fontMetrics = getFontMetrics(font2);
        axisTitleHeight = fontPixelSize2 + lineSpacing;
        if (!axisTitleTop.equals("")) {
            axisTitleTopWidth = fontMetrics.stringWidth(axisTitleTop);
            axisTitleTopHeight = axisTitleHeight;
        }
        if (!axisTitleBot.equals("")) {
            axisTitleBotWidth = fontMetrics.stringWidth(axisTitleBot);
            axisTitleBotHeight = axisTitleHeight;
        }
        if (!axisTitleLeft.equals("")) {
            axisTitleLeftWidth = fontMetrics.stringWidth(axisTitleLeft);
            axisTitleLeftHeight = axisTitleHeight;
        }
        if (!axisTitleRight.equals("")) {
            axisTitleRightWidth = fontMetrics.stringWidth(axisTitleRight);
            axisTitleRightHeight = axisTitleHeight;
        }

        fontMetrics = getFontMetrics(font1);
        valuesHeight = fontPixelSize1 + lineSpacing;
        leftPart = fontMetrics.stringWidth(".");
        if (displayRight) {
            valuesAreaDataWidthSwPix = fontMetrics.stringWidth("8") * valuesAreaDataWidthSW + fontMetrics.stringWidth("8") * roundScaleSW + leftPart;
        }
        if (displayLeft) {
            valuesAreaDataWidthNePix = fontMetrics.stringWidth("8") * valuesAreaDataWidthNE + fontMetrics.stringWidth("8") * roundScaleNE + leftPart;
        }

        d = getSize();
        if (d.width < 4 || d.height < 4) {
            throw new Error("Applet width and height must be at least 4 pixels each");
        }
        buffer = new BufferedImage((int) (d.width), (int) (d.height), BufferedImage.TYPE_INT_RGB);
        bufimg = buffer.createGraphics();

        //attach line indexes to SW or NE, if it is > -1
        Way = new String[inputDataCount];
        if (SWS) {
            for (int z = 0; z < southWest.length; z++) {
                if (southWest[z] < 0) {
                    continue;
                }
                Way[southWest[z]] = "SW";
            }
        }
        if (NES) {
            for (int z = 0; z < northEast.length; z++) {
                if (northEast[z] < 0) {
                    continue;
                }
                Way[northEast[z]] = "NE";
            }
        }
        //fill array with NaNs where indexes are =-1
        for (int z = 0; z < Way.length; z++) {
            if (Way[z] == null) {
                Way[z] = "NaN";
            }
        }

        legendSquareSize = fs1Unadopted;

        baseR = new BigDecimal[2][9];
        baseL = new BigDecimal[2][9];

        for (int q = 1; q <= 9; q++) {
            //base 1000
            baseR[0][q - 1] = BigDecimal.valueOf(1000d).pow(q);
            baseL[0][q - 1] = BigDecimal.ONE.divide(BigDecimal.valueOf(1000d).pow(q));
            //base 1024
            baseR[1][q - 1] = BigDecimal.valueOf(1024d).pow(q);
            baseL[1][q - 1] = BigDecimal.ONE.divide(BigDecimal.valueOf(1000d).pow(q));
        }
        
        formatterE = new DecimalFormat("Fexp");

        System.out.println("netZOND charting applet v 1.04");
        System.out.println("© johnrembo@zebra.lt");
        System.out.println("The source-code of netZOND applet comes under GPL2 License");
    }

    public synchronized void paint(Graphics g) {
        if (bufimg != null) {
            line("solid", 1);
            /*************************************************************LAYER-1: BORDER***********************************************/
            if (border > 0) {
                Rectangle L1 = new Rectangle(0, 0, d.width, d.height);
                GradientPaint G = new GradientPaint(d.width / 2, 0, Palette[colorIndex[0]], d.width / 2, d.height, Palette[colorIndex[1]], false);
                bufimg.setPaint(G);
                bufimg.fill(L1);
            }
            /********************************************************LAYER-2: GRAPH-TITLE AND BACKGROUND********************************/
            Rectangle L2 = new Rectangle(border, border, d.width - border * 2, d.height - border * 2);
            GradientPaint G = new GradientPaint((d.width - border) / 2, border, Palette[colorIndex[2]], (d.width - border) / 2, d.height - border, Palette[colorIndex[3]], false);
            bufimg.setPaint(G);
            bufimg.fill(L2);
            bufimg.setFont(font3);
            bufimg.setColor(Palette[colorIndex[4]]);
            if (!graphTitleTop.equals(null)) {
                bufimg.drawString(graphTitleTop, d.width / 2 - border * 2 - graphTitleTopWidth / 2, border + graphTitleTopHeight - fs3Bias);
            }
            if (!graphTitleBot.equals(null)) {
                bufimg.drawString(graphTitleBot, d.width / 2 - border * 2 - graphTitleBotWidth / 2, d.height - border - fs3Bias + lineSpacing);
            }
            /********************************************************LAYER-5: CHART FOREGROUND******************************************/
            bufimg.setFont(font1);
            fontMetrics = getFontMetrics(font1);
            valuesAreaTimeWidth = fontMetrics.stringWidth(currTime(0)) + 1;

            if (displayTop >= 0) {
                if (!rotatedTop) {
                    valuesAreaTopHeight = valuesHeight;
                } else {
                    if (scaleTop.toLowerCase().equals("data")) {
                        //jei traffic scale - butina numatyti vertes po kablelio, jei ne - atimam jau pridetas(init'e) vertes
                        if (Way[displayTop].equals("SW") && trafficScaleSW) {
                            valuesAreaTopHeight = valuesAreaDataWidthSwPix + lineSpacing + 1;
                        } else if (Way[displayTop].equals("SW") && !trafficScaleSW) {
                            valuesAreaTopHeight = valuesAreaDataWidthSwPix + lineSpacing + 1 + (fontMetrics.stringWidth("8") * roundScaleSW + leftPart);
                        } else if (Way[displayTop].equals("NE") && trafficScaleNE) {
                            valuesAreaTopHeight = valuesAreaDataWidthNePix + lineSpacing + 1;
                        } else if (Way[displayTop].equals("NE") && !trafficScaleNE) {
                            valuesAreaTopHeight = valuesAreaDataWidthNePix + lineSpacing + 1 + (fontMetrics.stringWidth("8") * roundScaleNE + leftPart);
                        }
                    } else {
                        valuesAreaTopHeight = valuesAreaTimeWidth + lineSpacing + 1;
                    }
                }
            } else {
                valuesAreaTopHeight = lineSpacing + 1;
            }

            if (displayBot >= 0) {
                if (!rotatedBot) {
                    valuesAreaBotHeight = valuesHeight;
                } else {
                    if (scaleBot.toLowerCase().equals("data")) {
                        if (Way[displayBot].equals("SW") && trafficScaleSW) {
                            valuesAreaBotHeight = valuesAreaDataWidthSwPix + lineSpacing + 1;
                        } else if (Way[displayBot].equals("SW") && !trafficScaleSW) {
                            valuesAreaBotHeight = valuesAreaDataWidthSwPix + lineSpacing + 1 + (fontMetrics.stringWidth("8") * roundScaleSW + leftPart);
                        } else if (Way[displayBot].equals("NE") && trafficScaleNE) {
                            valuesAreaBotHeight = valuesAreaDataWidthNePix + lineSpacing + 1;
                        } else if (Way[displayBot].equals("NE") && !trafficScaleNE) {
                            valuesAreaBotHeight = valuesAreaDataWidthNePix + lineSpacing + 1 + (fontMetrics.stringWidth("8") * roundScaleNE + leftPart);
                        }
                    } else {
                        valuesAreaBotHeight = valuesAreaTimeWidth + lineSpacing + 1;
                    }
                }
            } else {
                valuesAreaBotHeight = lineSpacing + 1;
            }

            marginLeft = border + axisTitleLeftHeight + valuesAreaDataWidthSwPix + lineSpacing + 1;
            marginRight = border + axisTitleRightHeight + valuesAreaDataWidthNePix + lineSpacing + 1;
            graphWidth = d.width - marginRight - marginLeft;

            //adopth graphwidth to gridlines.
            int graphWidthDelim = (graphWidth - chartInnerMarginVert) % step;
            graphWidth = graphWidth - graphWidthDelim;
            marginRight += graphWidthDelim;

            //determine legend count
            legendCount = 0;
            j = 0;
            for (i = 0; i < legendValue.length; i++) {
                if (!Way[i].equals("NaN") && !hideLine[i]) {
                    if (legendDetails[i] || !legendValue[i].equals("") || legendUnits[i] || currValue[i]) {
                        legendCount++;
                    }
                }//determine no empty legends
            }

            //determine which lines will contain legends
            legendIndexes = new int[legendCount];
            for (i = 0; i < legendValue.length; i++) {
                if (!Way[i].equals("NaN") && !hideLine[i]) {
                    if (legendDetails[i] || !legendValue[i].equals("") || legendUnits[i] || currValue[i]) {
                        legendIndexes[j] = i;
                        j++;
                    }
                }
                ;//put line with legends indexes to array 
            }

            legendAreaHeight = (legendCount / legendsPerLine + legendCount % legendsPerLine) * valuesHeight;
            //System.out.println("LH="+legendAreaHeight);
            marginTop = border + graphTitleTopHeight + axisTitleTopHeight + valuesAreaTopHeight;
            //System.out.println("marginTop=["+border+"]+["+graphTitleTopHeight+"]+["+axisTitleTopHeight+"]+["+valuesAreaTopHeight);
            marginBot = border + graphTitleBotHeight + valuesAreaBotHeight + axisTitleBotHeight + legendAreaHeight;
            //System.out.println("margintBot="+marginBot);
            graphHeight = d.height - marginBot - marginTop;
            //System.out.println("graphHeight=["+d.height+"]-["+marginBot+"]-["+marginTop+"]="+graphHeight);
            if (colorIndex.length > 13) {//if chart background color is set
                Rectangle RC = new Rectangle(marginLeft, marginTop, graphWidth, graphHeight);
                //bufimg.setColor(Palette[colorIndex[13]]);
                GradientPaint CG = new GradientPaint(marginLeft + (graphWidth / 2), d.height - marginBot, Palette[colorIndex[13]], marginLeft + (graphWidth / 2), marginTop, Palette[colorIndex[14]], false);
                bufimg.setPaint(CG);
                bufimg.fill(RC);
            }
            /*************************************************************LAYER-6: THE GRID*********************************************/
            //left right can't be written until we know exact size of graph to align to
            bufimg.setFont(font1);
            if (chartInnerMarginHorz > majHorzGridDensity) {
                chartInnerMarginHorz = chartInnerMarginHorz - majHorzGridDensity;
            }
            if (chartInnerMarginVert > majVertGridDensity) {
                chartInnerMarginVert = chartInnerMarginVert - majVertGridDensity;
            }
            int chartInnerMarginHorzMin = chartInnerMarginHorz;//can't be described before init
            int chartInnerMarginVertMin = chartInnerMarginVert;//can't be described before init
            if (minHorzGridDensity < chartInnerMarginHorz) {
                chartInnerMarginHorzMin = chartInnerMarginHorzMin - minHorzGridDensity;
            }
            if (minVertGridDensity < chartInnerMarginVert) {
                chartInnerMarginVertMin = chartInnerMarginVertMin - minHorzGridDensity;
            }
            majHorzGridLines = (graphHeight - chartInnerMarginHorz) / majHorzGridDensity;
            majVertGridLines = (graphWidth - chartInnerMarginVert) / majVertGridDensity;

            valuesLeft = new int[majHorzGridLines + 1];
            valuesBot = new int[majVertGridLines + 1];

            //minor vertical grid
            bufimg.setPaint(Palette[colorIndex[5]]);
            for (i = chartInnerMarginVertMin; i < graphWidth; i += minVertGridDensity) {
                bufimg.drawLine(marginLeft + i, d.height - marginBot, marginLeft + i, marginTop);
            }
            //minor horizontal grid
            for (i = chartInnerMarginHorzMin; i < graphHeight; i += minHorzGridDensity) {
                bufimg.drawLine(marginLeft, d.height - marginBot - i, marginLeft + graphWidth, d.height - marginBot - i);
                minHorzGridLines++;
            }
            //major vertical grid
            j = 0;
            bufimg.setPaint(Palette[colorIndex[6]]);
            for (i = chartInnerMarginVert; i <= graphWidth; i += majVertGridDensity) {
                bufimg.drawLine(X = marginLeft + i, d.height - marginBot, marginLeft + i, marginTop);
                valuesBot[j] = X;
                j++;
            }
            //major horizontal grid
            j = 0;
            for (i = chartInnerMarginHorz; i <= graphHeight; i += majHorzGridDensity) {
                bufimg.drawLine(marginLeft, Y = d.height - marginBot - i, marginLeft + graphWidth, d.height - marginBot - i);
                valuesLeft[j] = Y;
                j++;
            }

            /**********************************************************LAYER-7: CHART BORDER*******************************************/
            bufimg.setPaint(Palette[colorIndex[7]]);
            bufimg.drawRect(marginLeft, marginTop, graphWidth, graphHeight);
            /**************************************************************************************************************************/
            /*                                                 DATA GATHERING AND COLLECTING                                          */
            /**************************************************************************************************************************/
            if (collectionIndex == 0) {//NULL variables
                currentData = new double[20];
                currentData = fetch(path);
                //does script retur data count match config defined data acount?
                if (buff != inputDataCount) {
                    System.out.println("Script [" + path + "] returned incorrect data values count! (" + buff + " instead of " + inputDataCount + " records)");
                    return;
                } else {
                    System.out.println("Data received successfuly");
                }
                dataAmount = 1 + (graphWidth - chartInnerMarginHorz) / step;
                dataCollection = new double[buff][dataAmount];
                maxCollVal = new double[buff];
                minCollVal = new double[buff];
                //NULL array
                for (i = 0; i < buff; i++) {
                    dataCollection[i][collectionIndex] = currentData[i];//issaugom pirmas vertes masyve (dataIndex=0)
                    maxCollVal[i] = minCollVal[i] = currentData[i];

                    if (!Way[i].equals("NaN")) {
                        if (Way[i].compareTo("SW") == 0) {
                            southWestHorzMax = maxCollVal[i];
                            southWestHorzMin = minCollVal[i];
                        }
                    }
                    if (!Way[i].equals("NaN")) {
                        if (Way[i].compareTo("NE") == 0) {
                            northEastHorzMax = maxCollVal[i];
                            northEastHorzMin = minCollVal[i];
                        }
                    }
                    //update statistics arrays
                    if (legendDetails[i]) {
                        legendStatisticsMin[i] = String.valueOf(minCollVal[i]);
                        legendStatisticsMax[i] = String.valueOf(maxCollVal[i]);
                        legendStatisticsAvg[i] = String.valueOf(currentData[i]);
                    }
                }
                collectionIndex++;
            } else {//cumulated data array
                currentData = fetch(path);
                for (i = 0; i < buff; i++) {
                    //shifting
                    if (collectionIndex == dataAmount) {
                        // min max null values
                        if (dataCollection[i][1] >= 0) {
                            maxCollVal[i] = minCollVal[i] = currentData[i];
                        } else {
                            maxCollVal[i] = 0;
                            minCollVal[i] = currentData[i];
                        }
                        for (j = 0; j < collectionIndex - 1; j++) {
                            dataCollection[i][j] = dataCollection[i][j + 1];
                            if (dataCollection[i][j] > maxCollVal[i]) {
                                maxCollVal[i] = dataCollection[i][j];
                            }//min max values
                            if (dataCollection[i][j] < minCollVal[i]) {
                                minCollVal[i] = dataCollection[i][j];
                            }//min max values
                        }
                        shifted = true;
                    }
                    if (shifted) {
                        dataCollection[i][collectionIndex - 1] = currentData[i];
                    } else {
                        dataCollection[i][collectionIndex] = currentData[i];
                    }
                    if (currentData[i] > maxCollVal[i]) {
                        maxCollVal[i] = currentData[i];
                    }//min max values
                    if (currentData[i] < minCollVal[i]) {
                        minCollVal[i] = currentData[i];
                    }//min max values
                    if (!Way[i].equals("NaN")) {
                        if (Way[i].compareTo("SW") == 0) {
                            southWestHorzMax = maxCollVal[i];
                            southWestHorzMin = minCollVal[i];
                        }
                    }
                    if (!Way[i].equals("NaN")) {
                        if (Way[i].compareTo("NE") == 0) {
                            northEastHorzMax = maxCollVal[i];
                            northEastHorzMin = minCollVal[i];
                        }
                    }
                    if (legendDetails[i]) {//update statistics arrays
                        if (currentData[i] < Double.parseDouble(legendStatisticsMin[i])) {
                            legendStatisticsMin[i] = String.valueOf(currentData[i]);
                        }
                        if (currentData[i] > Double.parseDouble(legendStatisticsMax[i])) {
                            legendStatisticsMax[i] = String.valueOf(currentData[i]);
                        }
                        legendStatisticsAvg[i] = String.valueOf((Double.parseDouble(legendStatisticsAvg[i]) + currentData[i]) / 2);
                    }
                }
                if (!shifted) {
                    collectionIndex++;
                }
            }

            /*****************************************************STACK GRAPHS*******************************************************/
            for (i = 0; i < collectionIndex; i++) {

                if (SWS) {
                    for (j = 0; j < southWest.length; j++) {
                        if (southWest[j] < 0) {
                            continue;
                        }

                        if (!fixedScaleMaxSW && graphType[southWest[j]].toLowerCase().contains("stack") && j > 0) {
                            if (graphType[southWest[j - 1]].toLowerCase().contains("stack")) {
                                globalMaxSW += dataCollection[southWest[j]][i];
                            } else {
                                globalMaxSW += dataCollection[southWest[j]][i] + dataCollection[southWest[j - 1]][i];
                            }

                            if (globalMaxSW > southWestHorzMax) {
                                southWestHorzMax = globalMaxSW;
                            }
                        }
                    }
                }

                if (NES) {
                    for (j = 0; j < northEast.length; j++) {
                        if (northEast[j] < 0) {
                            continue;
                        }

                        if (!fixedScaleMaxNE && graphType[northEast[j]].toLowerCase().contains("stack") && j > 0) {
                            if (graphType[northEast[j - 1]].toLowerCase().contains("stack")) {
                                globalMaxNE += dataCollection[northEast[j]][i];
                            } else {
                                globalMaxNE += dataCollection[northEast[j]][i] + dataCollection[northEast[j - 1]][i];
                            }

                            if (globalMaxNE > northEastHorzMax) {
                                northEastHorzMax = globalMaxNE;
                            }
                        }
                    }
                }

                globalMaxSW = globalMaxNE = 0;
            }

            /****************************************************LAYER-8: AXES VALUES************************************************/
            axisValue = null;

            //horizontal grid            
            if (southWest != null) {
                for (i = 0; i < buff; i++) {
                    for (j = 0; j < southWest.length; j++) {
                        if (i == southWest[j]) {
                            if (Way[southWest[j]].equals("NaN")) {
                                continue;
                            } else {
                                horzScaleSW = true;
                            }
                            if (maxCollVal[i] > southWestHorzMax) {
                                southWestHorzMax = maxCollVal[i];
                            }
                            if (minCollVal[i] < southWestHorzMin) {
                                southWestHorzMin = minCollVal[i];
                            }
                        }
                    }
                }

                if (horzScaleSW) {
                    if (fixedScaleMinSW) {
                        southWestHorzMin = scaleMinSW;
                    }//manual min scale
                    if (fixedScaleMaxSW) {
                        southWestHorzMax = scaleMaxSW;
                    }//manual max scale
                    southWestHorzDelta = southWestHorzMax - southWestHorzMin;
                    if (southWestHorzDelta == 0) {
                        southWestHorzDelta = 1;
                    }//division by zero exeption

                    //axisValues calculation - left horizontal values
                    delta = new BigDecimal(southWestHorzDelta);
                    count = new BigDecimal(String.valueOf(valuesLeft.length - 1));//horizontal gridlines count
                    max = new BigDecimal(String.valueOf(southWestHorzMax));
                    delta = delta.divide(count, roundScaleSW, BigDecimal.ROUND_HALF_UP);
                    min = new BigDecimal(String.valueOf(southWestHorzMin));
                    
                    if (trafficScaleSW) {
                        direction = "SW";
                        factorSW = unitScale(southWestHorzMax);
                        roundScale = roundScaleSW;
                        delta = delta.divide(factorSW, roundScale, BigDecimal.ROUND_HALF_UP);
                        max = max.divide(factorSW, roundScale, BigDecimal.ROUND_HALF_UP);
                        min = min.divide(factorSW, roundScale, BigDecimal.ROUND_HALF_UP);
                    }

                    F = (trafficScale) ? ldSymbSW : Fexp;
                    //System.out.println("min="+min+" max="+max+" delta="+delta);                    
                    if (displayRight) {
                        if (retainAxisValuesColors && southWest.length == 1) {
                            gradientor(southWest[0], border, d.height - marginBot, marginLeft, marginTop, false);
                        } else {
                            bufimg.setColor(Palette[colorIndex[8]]);
                        }
                        if (delta.compareTo(zero) == 0 || min.equals(max)) {
                            axisValue = Format(min.toString(),F,valuesAreaDataWidthSW);
                            bufimg.drawString(axisValue, marginLeft - fontMetrics.stringWidth(String.valueOf(axisValue)), valuesLeft[0] + valuesHeight / 2);
                        } else {
                            for (i = 0; i < valuesLeft.length; i++) {
                                if (i != valuesLeft.length - 1) {
                                    axisValue = Format(min.toString(),F,valuesAreaDataWidthSW);
                                } else {
                                    axisValue = Format(max.toString(),F,valuesAreaDataWidthSW);
                                } //max scale value isn't allways correct (scale 0<->5 from 3 gridlines is made from: 0,1.666,3.33333,4.99999)
                                //is SW or NE under one line - paint left and right axis values
                                bufimg.drawString(axisValue, marginLeft - fontMetrics.stringWidth(axisValue), valuesLeft[i] + valuesHeight / 2);
                                min = min.add(delta);
                            }
                        }
                    }
                }
            //end axisValues calculation
            }

            if (northEast != null) {
                for (i = 0; i < buff; i++) {
                    for (j = 0; j < northEast.length; j++) {
                        if (i == northEast[j]) {
                            if (Way[northEast[j]].equals("NaN")) {
                                continue;
                            } else {
                                horzScaleNE = true;
                            }
                            if (maxCollVal[i] > northEastHorzMax) {
                                northEastHorzMax = maxCollVal[i];
                            }
                            if (minCollVal[i] < northEastHorzMin) {
                                northEastHorzMin = minCollVal[i];
                            }
                        }
                    }
                }
            
            if (horzScaleNE) {
                    if (fixedScaleMinNE) {
                        northEastHorzMin = scaleMinNE;
                    }
                    if (fixedScaleMaxNE) {
                        northEastHorzMax = scaleMaxNE;
                    }
                    northEastHorzDelta = northEastHorzMax - northEastHorzMin;
                    if (northEastHorzDelta == 0) {
                        northEastHorzDelta = 1;
                    }

                    //axisValues calculation - left horizontal values
                    delta = new BigDecimal(northEastHorzDelta);
                    count = new BigDecimal(String.valueOf(valuesLeft.length - 1));//horizontal gridlines count
                    max = new BigDecimal(String.valueOf(northEastHorzMax));
                    delta = delta.divide(count, roundScaleNE, BigDecimal.ROUND_HALF_UP);
                    min = new BigDecimal(String.valueOf(northEastHorzMin));
                    
                    if (trafficScaleNE) {
                        direction = "NE";
                        factorNE = unitScale(northEastHorzMax);
                        roundScale = roundScaleNE;
                        delta = delta.divide(factorNE, roundScale, BigDecimal.ROUND_HALF_UP);
                        max = max.divide(factorNE, roundScale, BigDecimal.ROUND_HALF_UP);
                        min = min.divide(factorNE, roundScale, BigDecimal.ROUND_HALF_UP);
                    }

                    F = (trafficScale) ? ldSymbNE : Fexp;
                    
                    if (displayLeft) {
                        if (retainAxisValuesColors && northEast.length == 1) {
                            gradientor(northEast[0], d.width - marginRight, d.height - marginBot, border, marginTop, false);
                        } else {
                            bufimg.setColor(Palette[colorIndex[8]]);
                        }
                        if (delta.compareTo(zero) == 0 || max.equals(min)) {
                            axisValue = Format(min.toString(),F,valuesAreaDataWidthNE);
                            bufimg.drawString(axisValue, d.width - marginRight + lineSpacing + 1, valuesLeft[0] + valuesHeight / 2);
                        } else {
                            for (i = 0; i < valuesLeft.length; i++) {
                                if (i != valuesLeft.length - 1) {
                                    axisValue = Format(min.toString(),F,valuesAreaDataWidthNE);
                                } else {
                                    axisValue = Format(max.toString(),F,valuesAreaDataWidthNE);
                                }
                                bufimg.drawString(axisValue, marginLeft + graphWidth + 2, valuesLeft[i] + valuesHeight / 2);//add 2 pix: 1 pix for border and 1 pix for space
                                min = min.add(delta);
                            }
                        }
                    }
                }
            }

            writeAxesLabels();//color 13
            //bufimg.setColor(Palette[colorIndex[8]]);//restore previous color
            //vertical grid
            for (i = 0; i < collectionIndex; i++) {
                for (j = 0; j < valuesBot.length; j++) {//bot or top - coordinates are neccesarry
                    if ((valuesBot[j] - marginLeft - chartInnerMarginVert) == i * step) {//matches every major-grid data value
                        //bottom axis values
                        if (displayBot >= 0 && !Way[displayBot].equals("NaN")) {
                            if (scaleBot.toLowerCase().equals("time")) {
                                tempValue = currTime(-(collectionIndex - i) * refreshInterval);
                            } else {
                                if (Way[displayBot].equals("SW")) {
                                    direction = "SW";
                                    tempValue = factoriaze(dataCollection[displayBot][i], factorSW);
                                } else if (Way[displayBot].equals("NE")) {
                                    direction = "NE";
                                    tempValue = factoriaze(dataCollection[displayBot][i], factorNE);
                                }
                            }

                            if (retainAxisValuesColors) {
                                bufimg.setColor(Palette[lineColor[displayBot][0]]);
                            } else {
                                bufimg.setColor(Palette[colorIndex[8]]);
                            }

                            if (rotatedBot) {
                                bufimg = rotate(bufimg, font1, tempValue, valuesBot[j] + fontSize1 / 2, d.height - marginBot + fontMetrics.stringWidth(tempValue) + lineSpacing, -90);
                            } else {
                                bufimg.drawString(tempValue, valuesBot[j] - fontMetrics.stringWidth(tempValue) / 2, d.height - marginBot + valuesHeight - lineSpacing);
                            }
                        }
                        //top axis values
                        if (displayTop >= 0 && !Way[displayTop].equals("NaN")) {
                            if (scaleTop.toLowerCase().equals("time")) {
                                tempValue = currTime(-(collectionIndex - i) * refreshInterval);
                            } else {
                                if (Way[displayTop].equals("NE")) {
                                    direction = "NE";
                                    tempValue = factoriaze(dataCollection[displayTop][i], factorNE);
                                } else if (Way[displayTop].equals("SW")) {
                                    direction = "SW";
                                    tempValue = factoriaze(dataCollection[displayTop][i], factorSW);
                                }
                            }

                            if (retainAxisValuesColors) {
                                bufimg.setColor(Palette[lineColor[displayTop][1]]);
                            } else {
                                bufimg.setColor(Palette[colorIndex[8]]);
                            }

                            if (rotatedTop) {
                                bufimg = rotate(bufimg, font1, tempValue, valuesBot[j] + fontSize1 / 2, d.height - marginBot - graphHeight - lineSpacing, -90);
                            } else {
                                bufimg.drawString(tempValue, valuesBot[j] - fontMetrics.stringWidth(tempValue) / 2, marginTop - lineSpacing);
                            }
                        }

                    }
                }
            }
            /*********************************************************LAYER-9: CHART***********************************************/
            String dataVal;
            bufimg.setColor(Palette[colorIndex[9]]);
            prevX = marginLeft + chartInnerMarginVert;

            Color[][] colArr;
            colArr = new Color[inputDataCount][collectionIndex];
            String[][] valArr;
            valArr = new String[inputDataCount][collectionIndex];
            xArr = new int[inputDataCount][collectionIndex];
            yArr = new int[inputDataCount][collectionIndex];

            for (i = 0; i < collectionIndex; i++) {//SW DATA AND PLOT
                X = marginLeft + chartInnerMarginVert + i * step;
                if (southWest != null) {
                    direction = "SW";
                    for (j = 0; j < southWest.length; j++) {
                        if (southWest[j] < 0) {
                            continue;
                        }
                        //j = jMin + (jMax - jMin)(y - yMax) / (yMin - yMax)
                        if (graphType[southWest[j]].toLowerCase().contains("stack") && j == 0) {
                            graphType[southWest[j]] = "plot";
                        }

                        if (!graphType[southWest[j]].toLowerCase().contains("stack")) {
                            if (dataCollection[southWest[j]][i] >= southWestHorzMax) {
                                Y = valuesLeft[valuesLeft.length - 1];
                            } else if (dataCollection[southWest[j]][i] <= southWestHorzMin) {
                                Y = valuesLeft[0];
                            } else {
                                Y = (int) Math.round((double) valuesLeft[valuesLeft.length - 1] +
                                        (double) (valuesLeft[0] - valuesLeft[valuesLeft.length - 1]) *
                                        (dataCollection[southWest[j]][i] - southWestHorzMax) /
                                        (southWestHorzMin - southWestHorzMax));
                            }
                        } else {
                            if (dataCollection[southWest[j]][i] + dataCollection[southWest[j - 1]][i] >= southWestHorzMax) {
                                Y = valuesLeft[valuesLeft.length - 1];
                            } else {
                                double Yprev = southWestHorzMax +
                                        (southWestHorzMin - southWestHorzMax) *
                                        ((double) prevYSW[j - 1] - (double) valuesLeft[valuesLeft.length - 1]) /
                                        (double) (valuesLeft[0] - valuesLeft[valuesLeft.length - 1]);

                                Y = (int) Math.round((double) valuesLeft[valuesLeft.length - 1] +
                                        (double) (valuesLeft[0] - valuesLeft[valuesLeft.length - 1]) *
                                        (dataCollection[southWest[j]][i] + Yprev - southWestHorzMax) /
                                        (southWestHorzMin - southWestHorzMax));
                            }
                        }
                        if (southWestHorzMax == southWestHorzMin) {
                            Y = valuesLeft[0];
                        }
                        if (prevYSW[j] == 0) {
                            prevYSW[j] = Y;
                        }

                        //data values bottom-left
                        dataVal = factoriaze(dataCollection[southWest[j]][i], factorSW);
                        //dataVal = String.valueOf(dataCollection[southWest[j]][i]);
                        line(lineTypes[southWest[j]], lineWidths[southWest[j]]);
                        if (!hideLine[southWest[j]]) {
                            drawObject(southWest[j], prevYSW[j]);
                        }
                        if (showValuesOnChart[southWest[j]] && !hideLine[southWest[j]]) {
                            colArr[southWest[j]][i] = bufimg.getColor();
                            valArr[southWest[j]][i] = dataVal;
                            xArr[southWest[j]][i] = X;
                        }
                        yArr[southWest[j]][i] = Y;
                        prevYSW[j] = Y;
                    }
                }
                if (northEast != null) {
                    direction = "NE";
                    for (j = 0; j < northEast.length; j++) {
                        if (northEast[j] < 0) {
                            continue;
                        }
                        //j = jMin + (jMax - jMin)(y - yMax) / (yMin - yMax)
                        if (graphType[northEast[j]].toLowerCase().contains("stack") && j == 0) {
                            graphType[northEast[j]] = "plot";
                        }

                        if (!graphType[northEast[j]].toLowerCase().contains("stack")) {
                            if (dataCollection[northEast[j]][i] >= northEastHorzMax) {
                                Y = valuesLeft[valuesLeft.length - 1];
                            } else if (dataCollection[northEast[j]][i] <= northEastHorzMin) {
                                Y = valuesLeft[0];
                            } else {
                                Y = (int) Math.round((double) valuesLeft[valuesLeft.length - 1] +
                                        (double) (valuesLeft[0] - valuesLeft[valuesLeft.length - 1]) *
                                        (dataCollection[northEast[j]][i] - northEastHorzMax) /
                                        (northEastHorzMin - northEastHorzMax));
                            }
                        } else {
                            if (dataCollection[northEast[j]][i] + dataCollection[northEast[j - 1]][i] >= northEastHorzMax) {
                                Y = valuesLeft[valuesLeft.length - 1];
                            } else {
                                double Yprev = northEastHorzMax +
                                        (northEastHorzMin - northEastHorzMax) *
                                        ((double) prevYNE[j - 1] - (double) valuesLeft[valuesLeft.length - 1]) /
                                        (double) (valuesLeft[0] - valuesLeft[valuesLeft.length - 1]);

                                Y = (int) Math.round((double) valuesLeft[valuesLeft.length - 1] +
                                        (double) (valuesLeft[0] - valuesLeft[valuesLeft.length - 1]) *
                                        (dataCollection[northEast[j]][i] + Yprev - northEastHorzMax) /
                                        (northEastHorzMin - northEastHorzMax));
                            }
                        }

                        if (northEastHorzMax == northEastHorzMin) {
                            Y = valuesLeft[0];
                        }
                        if (prevYNE[j] == 0) {
                            prevYNE[j] = Y;
                        }

                        //data values bottom-left
                        dataVal = factoriaze(dataCollection[northEast[j]][i], factorNE);
                        //dataVal = String.valueOf(dataCollection[northEast[j]][i]);
                        line(lineTypes[northEast[j]], lineWidths[northEast[j]]);
                        if (!hideLine[northEast[j]]) {
                            drawObject(northEast[j], prevYNE[j]);
                        }
                        if (showValuesOnChart[northEast[j]] && !hideLine[northEast[j]]) {
                            colArr[northEast[j]][i] = bufimg.getColor();
                            valArr[northEast[j]][i] = dataVal;
                            xArr[northEast[j]][i] = X;
                        }
                        yArr[northEast[j]][i] = Y;
                        prevYNE[j] = Y;
                    }
                }

                prevX = X;
            }
            //values on chart. Must be drawn on top of plot.
            for (i = 0; i < inputDataCount; i++) {
                for (j = 0; j < collectionIndex; j++) {
                    if (colArr[i][j] == null) {
                        continue;
                    }
                    //bufimg.setColor (colArr[i][j].darker ().darker ());
                    bufimg.setColor(colArr[i][j]);                    

                    if (SWS) {
                        if (inArray(southWest, i)) {
                            trafficScale = trafficScaleSW;
                            roundScale = roundScaleSW;
                            direction = "SW";
                            F = (trafficScale) ? ldSymbSW : Fexp;
                            bufimg.drawString(Format(String.valueOf(valArr[i][j]),F,valuesAreaDataWidthSW), xArr[i][j] - fontMetrics.stringWidth(valArr[i][j]) / 2, yArr[i][j] + fontSize1 / 2);                    
                        }
                    }

                    if (NES) {
                        if (inArray(northEast, i)) {
                            trafficScale = trafficScaleNE;
                            roundScale = roundScaleNE;
                            direction = "NE";
                            F = (trafficScale) ? ldSymbNE : Fexp;
                            bufimg.drawString(Format(String.valueOf(valArr[i][j]),F,valuesAreaDataWidthNE), xArr[i][j] - fontMetrics.stringWidth(valArr[i][j]) / 2, yArr[i][j] + fontSize1 / 2);                    
                        }
                    }
                    
                }
            }
            if (indicator && X < (d.width - marginRight)) {
                bufimg.setColor(Palette[colorIndex[10]]);
                line(indicatorType, indicatorWidth);
                bufimg.drawLine(X, marginTop + lineSpacing, X, d.height - marginBot - lineSpacing);
            }
            /********************************************************* DRAW ON SCREEN***********************************************/
            drawLegend();
            g.drawImage(buffer, 0, 0, this);
        }
    }

    public void start() {
        if (graphThread == null) {
            graphThread = new Thread(this, "graphThread");
            graphThread.start();
        }
    }

    public void stop() {
        if ((graphThread != null) && graphThread.isAlive()) //graphThread.stop () ;
        {
            graphThread = null;
        }
    }

    public void run() {
        while (true) {
            repaint();
            try {
                Thread.sleep(refreshInterval);
            } catch (InterruptedException e) {
            }
        }
    }

    public synchronized void update(Graphics g) {
        paint(g);
    }

    private Graphics2D rotate(Graphics2D bufimg, Font font, String text, int x, int y, int angle) {
        bufimg.translate(x, y);
        AffineTransform at = new AffineTransform();
        at.rotate(Math.toRadians(angle));
        bufimg.transform(at);
        Font font2 = font.deriveFont(new AffineTransform(1, 0, 0, 1, 0, 0));
        bufimg.setFont(font2);
        bufimg.drawString(text, 0, 0);
        bufimg.setTransform(new AffineTransform());
        bufimg.translate(0, 0);
        return bufimg;
    }

    public void line(String line_type, float line_width) {
        //default CAP_SQUARE,JOIN_MITER
        if (line_type.toLowerCase().equals("solid")) {
            bufimg.setStroke(new BasicStroke(line_width));
        } else if (line_type.toLowerCase().equals("dash")) {
            bufimg.setStroke(
                    new BasicStroke(
                    line_width,
                    BasicStroke.CAP_ROUND,
                    BasicStroke.JOIN_ROUND,
                    1f,
                    new float[]{2f},
                    0f));
        }
    }

    //exports data from within private function to class
    public void buffSave(int value) {
        buff = value;
    }

    //downloads data values from certain http page
    public double[] fetch(String address) {
        //System.out.println("opening address:"+address);
        String lineValue;
        double[] data;
        int lineCount = 0;

        data = new double[20];
        try {
            URL url = new URL(address);
            BufferedReader input = new BufferedReader(new InputStreamReader(url.openStream()));
            while ((lineValue = input.readLine()) != null && !lineValue.equals("")) {
                //System.out.println("red data value:"+lineValue);
                //if (lineValue.trim().matches("([0-9]+)")) {
                data[lineCount] = Double.parseDouble(lineValue);
                lineCount++;
            //} else {
            //bufimg.setColor(Color.RED);
            //bufimg.drawString(lineValue,(d.width-fontMetrics.stringWidth(lineValue))/2,(d.height-fontMetrics.getHeight())/2);                    
            //}
            }
            input.close();
            buffSave(lineCount);
        } catch (MalformedURLException e) {
            System.out.println(e);
        } catch (IOException e) {
            System.out.println(e);
        } catch (NumberFormatException e) {
            System.out.println(e);
        }
        return data;
    }

    public Color T(Color rgb, int transparencyIndex) {
        if (transparent) {
            rgb = new Color(rgb.getRed(), rgb.getGreen(), rgb.getBlue(), transparencyIndex);
        }
        return rgb;
    }

    public void drawObject(int obj, int prevY) {
        if (lineColor[obj][0] == lineColor[obj][1]) {
            bufimg.setColor(T(Palette[lineColor[obj][0]], transIndex[obj]));//if no color gradient
        } else {
            gradientor(obj, marginLeft, valuesLeft[0], marginLeft, valuesLeft[valuesLeft.length - 1], true);
        }

        if (graphType[obj].toLowerCase().equals("plot") && step == 1) {//1pix area-fill equals 1pix column mode.
            graphType[obj] = "bars";
        }

        if (!graphType[obj].toLowerCase().equals("bars")) {//columns can't be antialiassed
            if (lineSmothness[obj]) {
                bufimg.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
            }
            antiAlias = true;
        }

        if (Y == valuesLeft[valuesLeft.length - 1]) {
            Y++;
        }//do not allow line to be on top of upper gridline - lower by 1 pix.

        int[] pX, pY;

        /***********************BARS*******************/
        if (graphType[obj].toLowerCase().equals("bars")) {
            if (step == 1) {
                bufimg.drawLine(X, valuesLeft[0], X, Y);
            } else {
                if (prevX != X) {
                    //int[] pX,pY;
                    pX = new int[5];
                    pY = new int[5];
                    //bufimg.fillRect(prevX,Y,step,(int)pixCurr);
                    pX[0] = prevX;
                    pY[0] = valuesLeft[0] + 1;
                    pX[1] = prevX;
                    pY[1] = Y;
                    pX[2] = X;
                    pY[2] = Y;
                    pX[3] = X;
                    pY[3] = valuesLeft[0] + 1;
                    pX[4] = prevX;
                    pY[4] = valuesLeft[0] + 1;
                    bufimg.fillPolygon(pX, pY, pY.length);

                    if (lineContour[obj]) { //only for bars and stack-bars
                        if (prevX == (marginLeft + chartInnerMarginVert)) {
                            prevY = Y;
                        }
                        bufimg.setColor(T(Palette[lineColor[obj][1]].darker().darker(), transIndex[obj]));
                        if (prevY < Y) {
                            bufimg.drawLine(prevX - 1, prevY, prevX - 1, Y);
                        } else {
                            bufimg.drawLine(prevX, prevY, prevX, Y);
                        }
                        bufimg.drawLine(prevX, Y, X - 1, Y);
                    }

                }
            }
        /***********************LINE*******************/
        } else if (graphType[obj].toLowerCase().equals("line")) {
            if (prevX != X) {
                bufimg.drawLine(prevX, prevY, X, Y);
            }
        /***********************PLOT*******************/
        } else if (graphType[obj].toLowerCase().equals("plot")) {
            int[] Xtriangle, Ytriangle, Xsquare, Ysquare;
            Xtriangle = new int[3];
            Ytriangle = new int[3];
            Xsquare = new int[4];
            Ysquare = new int[4];
            if (prevY > Y) { //curve raises (coordinates are measured from top - this is why visual appearance is inverse)
                    /*       X                            Y         */
                Xtriangle[0] = prevX;
                Ytriangle[0] = prevY;
                Xtriangle[1] = X;
                Ytriangle[1] = Y;
                Xtriangle[2] = X;
                Ytriangle[2] = prevY;

                Xsquare[0] = prevX;
                Ysquare[0] = valuesLeft[0] + 1;
                Xsquare[1] = prevX;
                Ysquare[1] = prevY;
                Xsquare[2] = X;
                Ysquare[2] = prevY;
                Xsquare[3] = X;
                Ysquare[3] = valuesLeft[0] + 1;
            } else {//curve falls or remains the same
                Xtriangle[0] = prevX;
                Ytriangle[0] = prevY;
                Xtriangle[1] = X;
                Ytriangle[1] = Y;
                Xtriangle[2] = prevX;
                Ytriangle[2] = Y;

                Xsquare[0] = prevX;
                Ysquare[0] = valuesLeft[0] + 1;
                Xsquare[1] = prevX;
                Ysquare[1] = Y;
                Xsquare[2] = X;
                Ysquare[2] = Y;
                Xsquare[3] = X;
                Ysquare[3] = valuesLeft[0] + 1;
            }

            //int[] pX,pY; 
            pX = new int[7];
            pY = new int[7];
            pX[0] = Xsquare[0];
            pX[1] = Xsquare[1];
            pX[2] = Xtriangle[0];
            pX[3] = Xtriangle[1];
            pX[4] = Xtriangle[2];
            pX[5] = Xsquare[2];
            pX[6] = Xsquare[3];
            pY[0] = Ysquare[0];
            pY[1] = Ysquare[1];
            pY[2] = Ytriangle[0];
            pY[3] = Ytriangle[1];
            pY[4] = Ytriangle[2];
            pY[5] = Ysquare[2];
            pY[6] = Ysquare[3];
            if (prevX != X) {
                bufimg.fillPolygon(pX, pY, pY.length);
            }
        /*********************STACK-LINE*****************/
        } else if (graphType[obj].toLowerCase().equals("stack-line")) {

            int prevObjY, prevObjPrevY;

            if (direction.equals("SW")) {//sum of all SW plot and stack Y's
                prevObjY = (i > 0) ? yArr[southWest[j - 1]][i] : (valuesLeft[0] + 1);
                prevObjPrevY = (i > 0) ? yArr[southWest[j - 1]][i - 1] : (valuesLeft[0] + 1);
            } else {
                prevObjY = (i > 0) ? yArr[northEast[j - 1]][i] : (valuesLeft[0] + 1);
                prevObjPrevY = (i > 0) ? yArr[northEast[j - 1]][i - 1] : (valuesLeft[0] + 1);
            }

            if (prevX != X) {
                bufimg.drawLine(prevX, prevObjPrevY - (prevObjPrevY - prevY), X, prevObjY - (prevObjY - Y));
            }
        /*********************STACK-BARS*****************/
        //can't draw first stack sollumn if 1pix step chosen
        } else if (graphType[obj].toLowerCase().equals("stack-bars")) {//only if previously bars were drawn, otherwise - switch to stack-plot on line 851

            int prevObjY;
            boolean permitBars = false;

            if (direction.equals("SW")) {//sum of all SW plot and stack Y's
                prevObjY = (i > 0) ? yArr[southWest[j - 1]][i] : (valuesLeft[0] + 1);
                if (graphType[southWest[j - 1]].contains("bars")) {
                    permitBars = true;
                }
            } else {
                prevObjY = (i > 0) ? yArr[northEast[j - 1]][i] : (valuesLeft[0] + 1);
                if (graphType[northEast[j - 1]].contains("bars")) {
                    permitBars = true;
                }
            }
            if (prevX != X) {
                if (step == 1) {
                    if (prevObjY == Y) {
                        bufimg.drawLine(X, prevObjY, X, prevObjY - (prevObjY - Y));
                    } else {
                        bufimg.drawLine(X, prevObjY - 1, X, prevObjY - (prevObjY - Y));
                    }
                } else {
                    if (permitBars) {
                        //int[] pX,pY;
                        pX = new int[5];
                        pY = new int[5];

                        pX[0] = prevX;
                        pY[0] = prevObjY;
                        pX[1] = prevX;
                        pY[1] = prevObjY - (prevObjY - Y);
                        pX[2] = X;
                        pY[2] = prevObjY - (prevObjY - Y);
                        pX[3] = X;
                        pY[3] = prevObjY;
                        pX[4] = prevX;
                        pY[4] = prevObjY;
                        bufimg.fillPolygon(pX, pY, pY.length);

                        if (lineContour[obj]) { //only for bars and stack-bars
                            if (prevX == (marginLeft + chartInnerMarginVert)) {
                                prevY = Y;
                            }
                            bufimg.setColor(T(Palette[lineColor[obj][1]].darker().darker(), transIndex[obj]));
                            if (prevY < (prevObjY - (prevObjY - Y))) {
                                bufimg.drawLine(prevX - 1, prevY, prevX - 1, prevObjY - (prevObjY - Y));
                            } else {
                                bufimg.drawLine(prevX, prevY, prevX, prevObjY - (prevObjY - Y));
                            }
                            bufimg.drawLine(prevX, prevObjY - (prevObjY - Y), X - 1, prevObjY - (prevObjY - Y));
                        }


                    }
                }
            }
        /*********************STACK-PLOT*****************/
        } else if (graphType[obj].toLowerCase().equals("stack-plot")) {
            //int[] pX,pY;     
            pX = new int[5];
            pY = new int[5];
            int prevObjY, prevObjPrevY;

            if (direction.equals("SW")) {//sum of all SW plot and stack Y's
                prevObjY = (i > 0) ? yArr[southWest[j - 1]][i] : (valuesLeft[0] + 1);
                prevObjPrevY = (i > 0) ? yArr[southWest[j - 1]][i - 1] : (valuesLeft[0] + 1);
            } else {
                prevObjY = (i > 0) ? yArr[northEast[j - 1]][i] : (valuesLeft[0] + 1);
                prevObjPrevY = (i > 0) ? yArr[northEast[j - 1]][i - 1] : (valuesLeft[0] + 1);
            }
            pY[0] = prevObjPrevY;
            pX[0] = prevX;
            pY[1] = prevY;
            pX[1] = prevX;
            pY[2] = Y;
            pX[2] = X;
            pY[3] = prevObjY;
            pX[3] = X;
            pY[4] = prevObjPrevY;
            pX[4] = prevX;

            if (prevX != X) {
                bufimg.fillPolygon(pX, pY, pY.length);
            }

        } else {
            System.out.println("Unrecognized graph type:" + graphType[obj].trim());
        }

        if (lineContour[obj]) { //except for bars and stack-bars
            if (!(graphType[obj].toLowerCase().equals("stack-bars") || graphType[obj].toLowerCase().equals("bars"))) {
                bufimg.setColor(T(Palette[lineColor[obj][1]].darker().darker(), transIndex[obj]));
                if (prevX != X) {
                    bufimg.drawLine(prevX, prevY, X, Y);
                }
            }
        }

        if (antiAlias) {
            bufimg.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);
        }
    }

    public void drawLegend() {
        //legendCount contains record count per row, o legenValue contains records themselves
        int lastPos = 0;
        int currPosInLine = 1;
        String legendText = " ";
        String details = "";
        int legendX = 0;
        int legendY = 0;
        String currData = null;
        BigDecimal currScale = BigDecimal.ONE;
        int pixelsForThisLegend = graphWidth / legendsPerLine;
        boolean ex_curr,ex_min,ex_avg,ex_max;
        ex_curr=ex_min=ex_avg=ex_max=false;
        String legMin,legAvg,legMax;
        legMin=legAvg=legMax=null;
        
        if (legendY == 0) {
            legendY = marginTop + graphHeight + axisTitleBotHeight + valuesAreaBotHeight + valuesHeight - lineSpacing - 1;
        }//we can see font size as 11 pix, but measuring it visualy gives only 10 pix
        if (legendX == 0) {
            legendX = marginLeft;
        }
        for (i = 0; i < legendCount; i++) {//run through all the legends
            lastPos = legendIndexes[i];
            if (hideLine[lastPos]) {
                lastPos++;
                continue;
            }
            gradientor(lastPos, legendX, legendY, legendX, legendY - legendSquareSize, true);
            if (legendDetails[lastPos] || currValue[lastPos] || !legendValue[lastPos].equals("")) {
                bufimg.fillRect(legendX, legendY - legendSquareSize, legendSquareSize, legendSquareSize);
                currCol = bufimg.getColor();
                bufimg.setColor(Palette[colorIndex[15]]);
                bufimg.drawRect(legendX, legendY - legendSquareSize, legendSquareSize, legendSquareSize);
                bufimg.setColor(currCol);
            }

            if (!legendValue[lastPos].equals("")) {
                legendText = legendText + legendValue[lastPos];
            }
            if (!coloredLegendText) {
                bufimg.setColor(Palette[colorIndex[11]]);
            }

            if (trafficScale) {
                currScale = unitScale(maxCollVal[lastPos]);
                currData = factoriaze(dataCollection[lastPos][collectionIndex - 1], currScale);
            } else {
                currData = String.valueOf(dataCollection[lastPos][collectionIndex - 1]);
            }

            if (SWS) {
                if (inArray(southWest, lastPos)) {
                    trafficScale = trafficScaleSW;
                    roundScale = roundScaleSW;
                    direction = "SW";
                    F = (trafficScale) ? ldSymbSW : Fexp;
                }
            }

            if (NES) {
                if (inArray(northEast, lastPos)) {
                    trafficScale = trafficScaleNE;
                    roundScale = roundScaleNE;
                    direction = "NE";
                    F = (trafficScale) ? ldSymbNE : Fexp;
                }
            }

            if (legendDetails[lastPos]) {
                if (trafficScale) {
                    legMin=factoriaze(Double.parseDouble(legendStatisticsMin[lastPos]), currScale);
                    legAvg=factoriaze(Double.parseDouble(legendStatisticsAvg[lastPos]), currScale);
                    legMax=factoriaze(Double.parseDouble(legendStatisticsMax[lastPos]), currScale);
                  } else {
                    legMin=legendStatisticsMin[lastPos];
                    legAvg=legendStatisticsAvg[lastPos];
                    legMax=legendStatisticsMax[lastPos];
                }
            } else {
                details = "";
            }

            if (legendDetails[lastPos]) {
                if (SWS) {
                    details = " [" + Format(legMin,F,valuesAreaDataWidthSW) + "|" + Format(legAvg,F,valuesAreaDataWidthSW) + "|" + Format(legMax,F,valuesAreaDataWidthSW) + "]";
                        if (inArray(southWest, lastPos)) {
                            details += " " + (unitsSW!=null ? unitsSW : "");
                        }
                }
                if (NES) {
                    details = " [" + Format(legMin,F,valuesAreaDataWidthNE) + "|" + Format(legAvg,F,valuesAreaDataWidthNE) + "|" + Format(legMax,F,valuesAreaDataWidthNE) + "]";
                        if (inArray(northEast, lastPos)) {
                            details += " " + (unitsNE!=null ? unitsNE : "");
                        }
                    }
            }
            
            if (currValue[lastPos]) {
                if (SWS) {
                    if (inArray(southWest, lastPos)) {
                        legendText += ": " + Format(currData,F,valuesAreaDataWidthSW);
                    }
                }
                if (NES) {
                    if (inArray(northEast, lastPos)) {
                        legendText += ": " + Format(currData,F,valuesAreaDataWidthNE);
                    }
                }
            }
            
            bufimg.drawString(legendText + details, legendX + legendSquareSize, legendY);
            lastPos++;
            
            if (currPosInLine < legendsPerLine) {
                if (legendType.equals("static")) {
                    legendX += pixelsForThisLegend;
                }
                if (legendType.equals("dynamic")) {
                    legendX += fontMetrics.stringWidth(legendText) + legendSeparator + legendSquareSize + fontMetrics.stringWidth(details);
                }
                currPosInLine++;
            } else {
                legendY += valuesHeight;
                legendX = marginLeft;
                currPosInLine = 1;
            }
            legendText = " ";
        }
    }

    public String currTime(int reduce) {
        Date now = new Date(System.currentTimeMillis() + reduce);
        SimpleDateFormat sdf = new SimpleDateFormat(DATE_FORMAT);
        return sdf.format(now.getTime());
    }

    public BigDecimal shorten(String val) {
        BigDecimal currval = BigDecimal.valueOf(Double.parseDouble(val));  
        currval = currval.setScale(roundScale, BigDecimal.ROUND_HALF_UP);
        
        //check if we get 0.00. If it's a real 0 - return it, else - return full value 0.00000001 (otherwise setScale filters value)
        if (currval.toString().equals("0."+fill("","0",roundScale))) {
            return BigDecimal.valueOf(Double.valueOf(val));
        }

        return removeZeroes(currval);
    }
    
    public String Format(String Sval, String F, int len) {//Fact is necessary to show M, to big values are shown not converted to M scale
        NumberFormat Fr = new DecimalFormat(F);
        NumberFormat Fexp = new DecimalFormat("###E0");
        
        BigDecimal DCval = BigDecimal.valueOf(Double.parseDouble(Sval));
        String Les = DCval.setScale(roundScale, BigDecimal.ROUND_HALF_UP).toString();//Left shrinked String
        String Res = DCval.setScale(0, BigDecimal.ROUND_HALF_UP).toString();//Right shrinked String
        Double DBval = Double.valueOf(Sval);
        
        if (Double.valueOf(Les)<DBval) {// orig:0.000001, made:0.00, corrected: 1E-6
            if (BigDecimal.valueOf(DBval).compareTo(baseL[(dimensionFactor==1000 ? 0 : 1)][8])<0)
                return "0";
                Les=Fexp.format(Double.valueOf(Sval.toString()));
                return Les.toString().replace("E0", "");
        }
        
        if (Res.length()>len && Double.valueOf(Res)>1) {//orig:30000000, made:30000000, corrected: 30E6
            Res=Fexp.format(Double.valueOf(Sval.toString()));
            return Res.toString().replace("E0", "");
        }
        
        return removeZeroes(BigDecimal.valueOf(Double.parseDouble(Les))).toString();
    }

    public BigDecimal unitScale(double axisVal) {
        if (SWS && direction.equals("SW")) {
            dimensionFactor = dimensionFactorSW;
        }
        if (NES && direction.equals("NE")) {
            dimensionFactor = dimensionFactorNE;
        }
        int jump = 0;

        int base_index = 0;

        if (dimensionFactor == 1000) {
            base_index = 0;
        } else if (dimensionFactor == 1024) {
            base_index = 1;
        }

        if (axisVal >= dimensionFactor) { //1000b=1Kb
            while (axisVal >= dimensionFactor) {
                axisVal /= dimensionFactor;
                jump++;
            }
            
            if (jump < baseR[base_index].length) {
                tempString = shrtctRight[jump - 1];
                saveUnits();
                return baseR[base_index][jump - 1];
            } else {
                tempString = "";
                saveUnits();
                return BigDecimal.ONE;
            }

        } else if (axisVal < 1) { //0.9b = 900mb
            while (axisVal <= 1) {
                axisVal *= dimensionFactor;
                jump++;
            }
            if (jump < baseL[base_index].length) {
                tempString = shrtctLeft[jump - 1];
                saveUnits();
                return baseL[base_index][jump - 1];
            } else {
                tempString = "";
                saveUnits();
                return BigDecimal.ONE;
            }

        } else { // 1
            tempString = "";
            saveUnits();
            return BigDecimal.ONE;
        }

    }

    public void saveUnits() {
        if (SWS) {
            if (direction.equals("SW")) {
                unitsSW = dimensionOpen + tempString + dimensionSW + dimensionClose;
            }
        }
        if (NES) {
            if (direction.equals("NE")) {
                unitsNE = dimensionOpen + tempString + dimensionNE + dimensionClose;
            }
        }
    }

    public void writeAxesLabels() {//prints out LEFT/RIGHT axis labels atacching units if needed
        Font fontTemp = bufimg.getFont();
        bufimg.setFont(font2);
        bufimg.setColor(Palette[colorIndex[12]]);
        fontMetrics = getFontMetrics(font2);

        if (axisTitleTop != null) {
            if (!axisTitleTop.equals("")) {
                if (Way[displayTop].equals("SW") && trafficScaleSW) {
                    bufimg.drawString(axisTitleTop + unitsSW, d.width / 2 - border * 2 - (axisTitleTopWidth + fontMetrics.stringWidth(unitsNE)) / 2, border + graphTitleTopHeight + axisTitleTopHeight - fs2Bias - lineSpacing);
                } else if (Way[displayTop].equals("NE") && trafficScaleNE) {
                    bufimg.drawString(axisTitleTop + unitsNE, d.width / 2 - border * 2 - (axisTitleTopWidth + fontMetrics.stringWidth(unitsNE)) / 2, border + graphTitleTopHeight + axisTitleTopHeight - fs2Bias - lineSpacing);
                } else {
                    bufimg.drawString(axisTitleTop, d.width / 2 - border * 2 - (axisTitleTopWidth) / 2, border + graphTitleTopHeight + axisTitleTopHeight - fs2Bias - lineSpacing);
                }
            }
        }

        if (axisTitleBot != null) {
            if (!axisTitleBot.equals("")) {
                if (Way[displayBot].equals("SW") && trafficScaleSW) {
                    bufimg.drawString(axisTitleBot + unitsSW, d.width / 2 - border * 2 - (axisTitleBotWidth + fontMetrics.stringWidth(unitsSW)) / 2, d.height - border - graphTitleBotHeight - legendAreaHeight - fs2Bias);
                } else if (Way[displayBot].equals("NE") && trafficScaleNE) {
                    bufimg.drawString(axisTitleBot + unitsNE, d.width / 2 - border * 2 - (axisTitleBotWidth + fontMetrics.stringWidth(unitsNE)) / 2, d.height - border - graphTitleBotHeight - legendAreaHeight - fs2Bias);
                } else {
                    bufimg.drawString(axisTitleBot, d.width / 2 - border * 2 - axisTitleBotWidth / 2, d.height - border - graphTitleBotHeight - legendAreaHeight - fs2Bias);
                }
            }
        }

        if (axisTitleLeft != null) {
            if (!axisTitleLeft.equals("")) {
                if (trafficScaleSW) {
                    bufimg = rotate(bufimg, font2, axisTitleLeft + unitsSW, border + axisTitleLeftHeight - fs2Bias, marginTop + graphHeight / 2 + (axisTitleLeftWidth + fontMetrics.stringWidth(unitsSW)) / 2, -90);
                } else {
                    bufimg = rotate(bufimg, font2, axisTitleLeft, border + axisTitleLeftHeight - fs2Bias, marginTop + graphHeight / 2 + axisTitleLeftWidth / 2, -90);
                }
            }
        }

        if (axisTitleRight != null) {
            if (!axisTitleRight.equals("")) {
                if (trafficScaleNE) {
                    bufimg = rotate(bufimg, font2, axisTitleRight + unitsNE, d.width - border - axisTitleRightHeight + fs2Bias, marginTop + graphHeight / 2 - (axisTitleRightWidth + fontMetrics.stringWidth(unitsNE)) / 2, 90);
                } else {
                    bufimg = rotate(bufimg, font2, axisTitleRight, d.width - border - axisTitleRightHeight + fs2Bias, marginTop + graphHeight / 2 - axisTitleRightWidth / 2, 90);
                }
            }
        }

        bufimg.setFont(fontTemp);
        fontMetrics = getFontMetrics(fontTemp);
    }

    public String factoriaze(double orig, BigDecimal fact) {
        delta = new BigDecimal(String.valueOf(orig));

        if (SWS && direction.equals("SW")) {
            roundScale = roundScaleSW;
            dimensionFactor = dimensionFactorSW;
        }
        if (NES && direction.equals("NE")) {
            roundScale = roundScaleNE;
            dimensionFactor = dimensionFactorSW;
        }

        int base_index = 0;

        if (dimensionFactor == 1000) {
            base_index = 0;
        } else if (dimensionFactor == 1024) {
            base_index = 1;
        }

        if (delta.compareTo(baseR[base_index][baseR[base_index].length - 1]) >= 0) {
            return delta.toString();
        } else if (delta.compareTo(BigDecimal.ONE.divide(baseR[base_index][baseR[base_index].length - 1])) <= 0) {
            return delta.toString();
        }

        delta = delta.divide(fact, roundScale, BigDecimal.ROUND_HALF_UP);
        return String.valueOf(delta);
    }

    public BigDecimal removeZeroes(BigDecimal z) {
        for (int scale = 0;; scale++) {
            try {
                z = z.setScale(scale);
            } catch (ArithmeticException e) {
                continue;
            }

            return z;
        }
    }

    public void gradientor(int Pos, int x0, int y0, int x1, int y1, boolean transparency) {
        Color C1, C2;
        if (transparency) {
            C1 = T(Palette[lineColor[Pos][0]], transIndex[Pos]);
            C2 = T(Palette[lineColor[Pos][1]], transIndex[Pos]);
        } else {
            C1 = Palette[lineColor[Pos][0]];
            C2 = Palette[lineColor[Pos][1]];
        }
        if (lineColor[Pos][0] == lineColor[Pos][1]) {// if both colors equal - no gradient
            bufimg.setColor(C1);
        } else {
            GradientPaint P = new GradientPaint(x0, y0, C1, x1, y1, C2, true);
            bufimg.setPaint(P);
        }
    }

    public boolean inArray(int haystack[], int needle) {
        int l = haystack.length;
        if (l > 0) {
            for (int i = 0; i < l; i++) {
                if (haystack[i] == needle) {
                    return true;
                }
            }
            ;
        }
        return false;
    }

    //pixCurr = (dataCollection[northEast[j]][i]-northEastHorzMin)*pixHeight/(northEastHorzDelta);
    private double bias(double scaleRange, double valueRange) {
        int remainder;
        double tempResult;

        tempResult = scaleRange / valueRange;
        remainder = (int) scaleRange % (int) valueRange;
        if (remainder == 0) {
            return 0;
        }
        valueRange++;
        return tempResult / (scaleRange / valueRange);
    }
    
    private String fill (String s, String f, int n) {
        if (n<=0)
            return s;
        
        while (n>0) {
            s+=f;
        n--;
        }
        
        return s;
    }
}